/* ser_sr24.h
 * 
 * Provides constants and declarations for the 
 * model SR24 device.
 * 
 * Copyright (C) 2000, Winford Engineering.
 * See file LICENSE.txt for details.
 * 
 * Winford Engineering
 * www.winfordeng.com
 */
#ifndef ser_sr24_h
#define ser_sr24_h

#include "ser_serial.h"
#include "ser_threads.h"

#ifndef COMMONWIN
#ifdef WIN95
#define COMMONWIN
#endif
#ifdef WINNT
#define COMMONWIN
#endif
#endif

#ifdef COMMONWIN
#define CALLCONVENTION __stdcall
#endif

#ifdef LINUX
#define CALLCONVENTION
#endif

/* First, private constants, etc.  These items are only used internally,
 * and are useless to the end user, so they don't have to be included
 * in the release header, but it doesn't hurt if they're left in either
 */ 

/* range of port numbers: */
#define SER_SR24_PORT_MIN	1
#define SER_SR24_PORT_MAX	3
/* digital bit numbers: */
#define SER_SR24_BIT_MIN	0
#define SER_SR24_BIT_MAX	7
/* analog channel numbers: */
#define SER_SR24_CHAN_MIN   0
#define SER_SR24_CHAN_MAX   7

#define SER_SR24_MAX_PROPERTY 2


/* Format for callback function, used with events */
typedef void (CALLCONVENTION *ser_eventfn)(int id, int value, void *extra);

/* structure used to specify how events should be handled */
typedef struct
{
	int type; /* which method to use for event notification */
	          /* 0 - none; disabled.
	           * 1 - callback function
	           * 2 - Windows message: sent to window ID w/ msg id
	           *     specified.  wparam will be the event id and
	           *     lparam will be the state (value).
	           * 3 - special callback, uses windows messages.
	           *     In this mode, the API creates its own window
	           *     and message processing function for that window,
	           *     and depends on the application to keep up with the
	           *     event loop.  The API's message processing function
	           *     then receives a pointer to data describing the event
	           *     and makes a callback to the application with that
	           *     data.  This has the advantage of being compatible with
	           *     Visual Basic because everything is done in VB's own
	           *     thread.
	           */
	ser_eventfn eventfn; /* used if type is 1 or 3. points to user
	                                     * defined function.
	                                     */
	void *window; /* if type is 2: window handle to send a message to */
	unsigned int msgid; /* If type is 2: message id to be sent */
	void *extra; /* If type is 1 or 3: extra data to be passed to the callback. */
} ser_event_spec;


#ifdef COMMONWIN
#define PIPE_BUFFER 4096

#endif

/* these are the powers of the bitmasks for the mode member of the structure */
#define SR24_BUFFER     0 /* on for enabled, off for disabled */
#define SR24_DIRECTION  1 /* on for output, off for input */
#define SR24_TYPE       2 /* on for analog, off for digital */

/* Serial Baud Rates: ----------------------------------------------- */

#define SER_1200   9
#define SER_2400  11
#define SER_4800  12
#define SER_9600  13
#define SER_19200 14
#define SER_38400 15

/* The following items are a bit more specific to this model board */

/* these constants are for the mode of set_portmode function */
#define SR24_BUFFER_ENABLED          0
#define SR24_BUFFER_DISABLED         1
#define SR24_DIRECTION_INPUT         0
#define SR24_DIRECTION_OUTPUT        1
#define SR24_TYPE_DIGITAL            0
#define SR24_TYPE_ANALOG             1

typedef struct
{
	int quitflag; /* normally 0, 1 to indicate that the device should be closed. 
					Then, the reading thread will set it back to 0 when it 
					recognizes that it should exit. */
	int event_quitflag; /* same situation as quitflag, but this is for the event thread,
	                     * although the event thread recognizes that it should exit
	                     * because the pipe has been closed 
	                     */
	int verifyflag; /* normally 0, 1 to indicate that the device is being verified.
	                 * When the reading thread gets a ping response, it will set
	                 * this back to 0 if it is 1.
	                 */
	int mode[3]; /* indicates mode of each port */
	int portclean; /* provides bitmask of which ports are clean. 1=clean */
	int portvalue[3]; /* most recent value of port. */
	int bitclean[3]; /* indicates which bits are clean.  one int per port. each int is a
	                  * bitmask indicating which bits are clean. 1=clean. */
	int bitvalue[3]; /* most recent bit values. one int per port. each int is a bitmask */
	int analogclean; /* bitmask of which analog values are clean 1=clean */
	int analogvalue[8]; /* most recent analog channel values */

					  /* Event IDs: The events on each bit and port may be assigned
					   * an id.  An id of 0 means the event is not wanted.
					   * Only positive IDs may be assigned. Negative
					   * IDs are reserved for error codes
					   */
	int portevent[3]; /* holds event id's for port events. 0 if not wanted */
	int bitevent[3][8]; /*holds event id's for each bit. 0 if not wanted */
	
	unsigned int timeout; /* timeout for user requests such as input (in milliseconds) 
	                         0 means infinite timeout */
	
	ser_serial serial; /* Handle of the serial port we're using */
	ser_mutex mutex;   /* mutex used to protect this structure's data 
	                      This protects all data that isn't already protected
	                      with an event predicate mutex or something */
	ser_event event;   /* event used to signal when new data arrives */

	/* these variables are intended to handle synchronous data events */
	ser_event synch_event; /* whenever an [data event from device] occurs, a 
	                            * signal will be broadcast on this event. */
	int last_event; /* the id of the last event that occurred */
	int last_event_value; /*the state of the event specified by last_event */
	
	/* data events */
	ser_event_spec event_handler;	
	
	/* if the reader thread gets an event and the bitevent or portevent
	 * field is set to zero, the event will be ignored
	 */

#ifdef LINUX
	/* If we are on Linux, we use a third thread to execute callback events,
	 * and we use a pipe to send the event specifications to that third thread.
	 * The pipe also acts as a buffer, allowing the events to be queued up
	 * in case the callback function takes a while to execute.
	 */
	 int pipe_event[2];  /* filedescriptors of the pipe.
	                      * [0] is for reading by the event handler thread
	                      * [1] is for writing by the serial-port-reading thread
	                      */
#endif

#ifdef COMMONWIN
	HANDLE pipe_event[2]; /* handles for pipe.  Pipe's purpose is same as in Linux. see above 
	                       * As on linux:
	                       * [0] is for reading by event handler thread
	                       * [1] is for writing
	                       */

	HANDLE wincb; /* handle to window for callbacks if event handler type 3 is being used */
	/* handles to threads.  These are created by the threads themselves
	 * right at the beginning of the function.  These are used
	 * to terminate the thread if that is necessary at Dll unload
	 * time
	 */
	HANDLE evthread_handle;   
	HANDLE readthread_handle; 
#endif

} ser_sr24_data;


/* define the ser_sr24_eventblock structure.
 * This is the format for the events as they are pushed through
 * the pipe to the event handler thread.
 */
typedef struct
{
	int id; /* Id of the event that occurred */
	int value; /* value that the specified event ID now has */
	void *extra; /* extra data for function */
} ser_sr24_eventblock;

/* Functions: the core of the API */
ser_sr24_data* CALLCONVENTION ser_sr24_open(char *serial, unsigned int baud, int *result);
int CALLCONVENTION ser_sr24_verify_connection(ser_sr24_data *data);
int CALLCONVENTION ser_sr24_reset_board(ser_sr24_data *data);
int CALLCONVENTION ser_sr24_output_byte(ser_sr24_data *data, int port, int value);
int CALLCONVENTION ser_sr24_input_byte(ser_sr24_data *data, int port, int *value);
int CALLCONVENTION ser_sr24_output_bit(ser_sr24_data *data, int port, int bit, int value);
int CALLCONVENTION ser_sr24_input_bit(ser_sr24_data *data, int port, int bit, int *value);
int CALLCONVENTION ser_sr24_input_analog(ser_sr24_data *data, int channel, int *value);
int CALLCONVENTION ser_sr24_get_portmode(ser_sr24_data *data, int port, int property);
int CALLCONVENTION ser_sr24_set_portmode(ser_sr24_data *data, int port, int property, int mode);
int CALLCONVENTION ser_sr24_set_timeout(ser_sr24_data *data, unsigned int timeout);
int CALLCONVENTION ser_sr24_register_event(ser_sr24_data *data, int port, int bit, int id);
int CALLCONVENTION ser_sr24_set_event_handler(ser_sr24_data *data, ser_event_spec *event);
int CALLCONVENTION ser_sr24_wait_event(ser_sr24_data *data, int *events, unsigned int length, unsigned int timeout, int *value);
int CALLCONVENTION ser_sr24_close(ser_sr24_data *data);
char* CALLCONVENTION ser_sr24_error_string(int errorcode);

#endif /* ser_sr24_h */
